// explore the GP code base

defineClass ClassBrowser morph window classes classesFrame scripts scriptsFrame fields fieldsFrame methods methodsFrame currentFunction editedFunction wasEdited

to newClassBrowser {
  sys = (new 'ClassBrowser')
  initialize sys
  return sys
}

method initialize ClassBrowser {
  scale = (global 'scale')
  window = (window 'Class Browser')
  clr = (clientColor window)
  bg = (gray 240)
  morph = (morph window)
  setHandler morph this
  setMinExtent morph (scale * 300) (scale * 200)

  classes = (listBox (sortedClasses this) (action 'classListName' this)  (action 'selectClass' this) clr (action 'openSynopsis' this))
  classesFrame = (scrollFrame classes clr)
  addPart morph (morph classesFrame)

  fields = (listBox (array) (action 'fieldRepresentation' this) (action 'selectField' this) bg nil nil 230)
  fieldsFrame = (scrollFrame fields bg)
  addPart morph (morph fieldsFrame)

  methods = (listBox (array) 'blockForFunction' (action 'selectMethod' this) bg (action 'blockify' this) nil 230)
  methodsFrame = (scrollFrame methods bg)
  addPart morph (morph methodsFrame)

  scripts = (newScriptEditor 10 10)
  scriptsFrame = (scrollFrame scripts bg)
  addPart morph (morph scriptsFrame)

  wasEdited = false

  setExtent morph (scale * 400) (scale * 300)
}

method globalBlocksName ClassBrowser {return '<Global Blocks>'}

method classListName ClassBrowser aClass {
  if (aClass == (globalBlocksName this)) {return aClass}
  return (className aClass)
}

method sortedClasses ClassBrowser {
  return (join
    (array (globalBlocksName this))
    (sorted (classes) (function c1 c2 { return ((className c1) < (className c2)) }))
  )
}

method sortedFunctions ClassBrowser {
  return (sorted (functions (topLevelModule)) (function f1 f2 { return ((functionName f1) < (functionName f2)) }))
}

method fixLayout ClassBrowser {
  classesWidth = (action 'spaceBoundedBy' window (action 'allWidth' classes) 25)
  fieldsWidth = (action 'spaceBoundedBy' window  150 25)
  methodsWidth = (action 'width' (morph fieldsFrame))

  fixLayout window
  packer = (newPanePacker (clientArea window) ((border window) / (global 'scale')))
  packPanesH packer classesFrame classesWidth fieldsFrame fieldsWidth scriptsFrame 'rest'
  packPanesH packer classesFrame 'done' methodsFrame methodsWidth
  packPanesV packer classesFrame '100%'
  packPanesV packer fieldsFrame '30%' methodsFrame 'rest'
  packPanesV packer scriptsFrame '100%'
  finishPacking packer
}

method redraw ClassBrowser {
  redraw window
  fixLayout this
}

method blockify ClassBrowser aFunction {
  page = (page morph)
  if (isNil (cmdList aFunction)) { return } // empty function
  block = (toBlock (cmdList aFunction))
  setPosition (morph block) (x (hand page)) (y (hand page))
  addPart page block
}

method browse ClassBrowser aClass aMethod {
  select classes aClass
  scrollIntoView classesFrame (bounds (selectedMorph classes))
  if (isNil aMethod) {return}
  select methods aMethod
  scrollIntoView methodsFrame (bounds (selectedMorph methods))
}

method selectClass ClassBrowser aClass {
  if (aClass == (globalBlocksName this)) {
    setCollection fields (variableNames (sessionModule))
  } else {
    setCollection fields (fieldNames aClass)
  }
  setCollection methods (sortedMethods this aClass)
}

method selectField ClassBrowser fieldName {
  if (shiftKeyDown (keyboard (handler (root morph)))) {
    unselectFields this
    return
  }
  setCollection methods (accessors this fieldName)
}

method unselectFields ClassBrowser {
  select fields nil
  cls = (selection classes)
  mth = (array)
  if (notNil cls) {mth = (sortedMethods this cls)}
  setCollection methods mth
}

method accessors ClassBrowser fieldName {
  result = (list)
  for m (sortedMethods this (selection classes)) {
    if (contains (varsUsed m) fieldName) {add result m}
  }
  return (toArray result)
}

method fieldRepresentation ClassBrowser fieldName {
  if (== (selection classes) (globalBlocksName this)) {
    return (newReporter 'global' fieldName)
  }
  return (newReporter 'my' fieldName)
}

method sortedMethods ClassBrowser aClass {
  if (aClass == (globalBlocksName this)) {return (sortedFunctions this)}
  return (sorted (methods aClass) (function f1 f2 { return ((functionName f1) < (functionName f2)) }))
}

method selectMethod ClassBrowser aFunction {
  if wasEdited {
  msg = 'This function has been edited.
Save it?'
  if (confirm (page morph) 'Save Changes?' msg) {
      saveEditedFunction this
    }
  }
  currentFunction = aFunction
  editedFunction = (copy aFunction)
  wasEdited = false
  refreshEditedFunction this
}

method refreshEditedFunction ClassBrowser {
  removeAllParts (morph scripts)
  if (isNil editedFunction) {return}
  block = (scriptForFunction editedFunction)
  setGrabRule (morph block)'ignore'
  setPosition (morph block) (left (morph scripts)) (top (morph scripts))
  off = ((global 'scale') * 10)
  moveBy (morph block) off off
  addPart (morph scripts) (morph block)
  frame = (handler (owner (morph scripts)))
  setPosition (morph scripts) (left (morph frame)) (top (morph frame))
  updateSliders frame
}

method openSynopsis ClassBrowser aClass {
  page = (page morph)
  syn = (newSynopsis)
  setPosition (morph syn) (x (hand page)) (y (hand page))
  addPart page syn
  browse syn aClass
}

// context menus

method rightClicked ClassBrowser hand {
  popUpAtHand (contextMenu this) (page hand)
  return true
}

method handleContextRequest ClassBrowser origin {
  if (origin === fieldsFrame) {
    menu = (fieldListContextMenu this)
  } (origin === methodsFrame) {
    menu = (methodListContextMenu this)
  }
  if (notNil menu) {
    popUpAtHand menu (global 'page')
  }
}

method handleListContextRequest ClassBrowser anArray {
  origin = (first anArray)
  dta = (data (last anArray))
  if (origin === classes) {
    menu = (classContextMenu this dta)
  } (origin === fields) {
    menu = (fieldContextMenu this dta)
  } (origin === methods) {
    menu = (methodContextMenu this dta)
  }
  if (notNil menu) {
    popUpAtHand menu (global 'page')
  }
}

method contextMenu ClassBrowser {
  menu = (menu nil this)
  addItem menu 'Make a new Class...' 'useScripter'
  return menu
}

method classContextMenu ClassBrowser aClass {
  menu = (menu (join 'Class:' (newline) (className aClass)) this)
  addItem menu 'delete...' 'useScripter'
  addItem menu 'rename...' 'useScripter'
  addLine menu
  addItem menu 'blocks palette...' (action 'openBlocksPalette' this aClass)
  addItem menu 'browse...' (action 'browseClass' aClass) 'open in a new window'
  addLine menu
  addItem menu 'make a new class...' 'useScripter'
  return menu
}

method fieldListContextMenu ClassBrowser {
  cls = (selection classes)
  menu = (menu nil this)
  if (notNil cls) {
    addItem menu 'make a field...' 'useScripter' (join 'add a new field to class ' (classListName this cls))
  }
  if (notNil (selection fields)) {
    addLine menu

    addItem menu 'unselect' (action 'unselectFields' this)
  }
  addItem menu 'make a new class...' 'useScripter'
  return menu
}

method fieldContextMenu ClassBrowser fieldName {
  menu = (menu nil this)
  addBlock this (toBlock (fieldRepresentation this fieldName)) menu
  addItem menu 'delete...' 'useScripter'
  addItem menu 'rename...' 'useScripter'
  addLine menu
  if (notNil (selection fields)) {
    addItem menu 'unselect' (action 'unselectFields' this)
  }
  if (notNil (selection classes)) {
    addItem menu 'make a field...' 'useScripter'
  }
  return menu
}

method methodListContextMenu ClassBrowser {
  cls = (selection classes)
  menu = (menu nil this)
  if (isNil cls) {
    addItem menu 'make a new class...' 'useScripter'
  } else {
    addItem menu 'make a block...' 'makeNewBlock' (join 'add a new method to class ' (classListName this cls))
  }
  return menu
}

method methodContextMenu ClassBrowser aFunction {
  menu = (menu nil this)
  addBlock this (blockForFunction aFunction) menu
  addItem menu 'delete...' (action 'deleteFunction' this aFunction (selection classes))
  addItem menu 'rename...' 'useScripter'
  addItem menu 'other implementations...' (action 'browseImplementors' this aFunction (selection classes))
  if (notNil (selection classes)) {
    addLine menu
    addItem menu 'make a block...' 'makeNewBlock' (join 'add a new method to class ' (classListName this (selection classes)))
  }
  return menu
}

to browseClass aClass methodName {
  page = (global 'page')
  brs = (newClassBrowser)
  setPosition (morph brs) (x (hand page)) (y (hand page))
  addPart page brs
  if (== aClass (globalBlocksName brs)) {
    browse brs aClass (functionNamed methodName (topLevelModule))
  } else {
    browse brs aClass (methodNamed aClass methodName)
  }
}

method openBlocksPalette ClassBrowser aClass {
  pal = (openBlocksPalette (global 'page'))
  browse pal aClass
}

to useScripter {
  inform (global 'page') (join 'under construction' (newline) 'please use Scripter')
}

method addBlock ClassBrowser aBlock menu {
  addItem menu (fullCostume (morph aBlock)) (action 'grab' this aBlock) 'pick up this block and use it in a script'
}

method grab ClassBrowser aBlock {
  h = (hand (global 'page'))
  setCenter (morph aBlock) (x h) (y h)
  grab h aBlock
}

method makeNewBlock ClassBrowser {
  page = (global 'page')
  cls = (selection classes)
  if (isNil cls) {return}
  name = (prompt page 'Enter a new block name:' 'myBlock')
  if (name == '') {return}
  if ((selection classes) == (globalBlocksName this)) {
    for f (functions (topLevelModule)) {
      if (== name (functionName f)) {
        inform page (join 'a global block named' (newline) name (newline)'already exists')
        return
      }
    }
    func = (defineFunctionInModule (topLevelModule) name)
    setCollection methods (sortedFunctions this)
  } else {
    for f (methods cls) {
      if (== name (functionName f)) {
        inform page (join 'a block named' (newline) name (newline) 'already exists in class' (newline) (className cls))
        return
      }
    }
	func = (addMethod cls name)
	spec = (blockSpecFromStrings name ' ' (join name ' _') (className cls))
    setCollection methods (sortedMethods this cls)
  }
  select methods func
  scrollIntoView methodsFrame (bounds (selectedMorph methods))
}

method deleteFunction ClassBrowser method class {
  if ((selection classes) == (globalBlocksName this)) {
    if (confirm (global 'page') nil (join 'Are you sure you want to delete the global block named' (newline) (functionName method) '?')) {
      remove (globalFuncs) (functionName method)
      if (method === (selection methods)) {
        select methods nil
        selectMethod this nil
      }
      setCollection methods (sortedMethods this class)
    }
  } else {
    if (confirm (global 'page') nil (join 'Are you sure you want to delete the block named' (newline) (functionName method) (newline) (join 'from class ' (classListName this class) '?'))) {
      removeMethodNamed class (functionName method)
      if (method === (selection methods)) {
        select methods nil
        selectMethod this nil
      }
      setCollection methods (sortedMethods this class)
    }
  }
}

method browseImplementors ClassBrowser method class {
  name = (functionName method)
  implementors = (implementors name)
  if (isEmpty implementors) {
    inform (global 'page') (join name (newline) 'is a primitive')
    return
  }
  others = (list)
  selectedClassName = (className (selection classes))
  for each implementors {
    if (!= each selectedClassName) {add others each}
  }
  if (isEmpty others) {
    inform (global 'page') (join name (newline) 'is not implemented' (newline) 'anywhere else')
    return
  }
  menu = (menu (join 'other implementations of' (newline) name) (action 'openClassBrowser' this name) true) // reverse call
  for each others {
    addItem menu (join each '...') each
  }
  popUpAtHand menu (global 'page')
}

method openClassBrowser ClassBrowser functionName aClassName {
  if (== aClassName '<generic>') {
    browseClass '<Global Blocks>' functionName
  } else {
    browseClass (class aClassName) functionName
  }
}

// reflect user edits

method wasEdited ClassBrowser {return wasEdited}

method blockPrototypeChanged ClassBrowser {
  wasEdited = true
  refreshEditedFunction this
}

method functionTypeChanged ClassBrowser {
  wasEdited = true
  refreshEditedFunction this
}

method functionBodyChanged ClassBrowser {
  wasEdited = true
  refreshEditedFunction this
}

method saveEditedFunction ClassBrowser {
  setField currentFunction 'argNames' (argNames editedFunction)
  updateCmdList currentFunction (cmdList editedFunction)
  editedFunction = (copy currentFunction)
  wasEdited = false
  for item (parts (morph methods)) {
    meth = (data (handler item))
    if (meth === currentFunction) {
      (replaceCostumes (handler item) nil nil nil)
      refresh (handler item)
      if (currentfunction === (selection methods)) {
        selectMethod this currentFunction
      }
    }
  }
}

// serialization

method preSerialize ClassBrowser {
  setCollection classes (array)
  preSerialize scripts
}

method postSerialize ClassBrowser {
   setCollection classes (sortedClasses this)
  postSerialize scripts
}
