// a self-updating thumbnail of a morphic handler

defineClass SpriteThumbnail morph target

to thumbnailFor aHandler {
  if (isClass aHandler 'Morph') {aHandler = (handler aHandler)}
  return (initialize (new 'SpriteThumbnail') aHandler)
}

method initialize SpriteThumbnail aHandler {
  target = aHandler
  morph = (newMorph this)
  setTransparentTouch morph true
  setWidth (bounds morph) (40 * (global 'scale'))
  setHeight (bounds morph) (40 * (global 'scale'))
  step this
  return this
}

method step SpriteThumbnail {
  if (isNil target) {return}
  targetM = (morph target)
  if (and
	((costumeData morph) === (costumeData targetM))
	((rotation morph) == (rotation targetM))) {
	  return
  }

  // determine minimum scale (keep aspect ratio)
  if (((width targetM) * (scaleX targetM)) == 0) {
	xRatio = 1
  } else {
	xRatio = ((width (bounds morph)) / ((width targetM) / (scaleX targetM)))
  }
  if (((height targetM) * (scaleY targetM)) == 0) {
	yRatio = 1
  } else {
	yRatio = ((height (bounds morph)) / ((height targetM) / (scaleY targetM)))
  }
  newScale = (min xRatio yRatio)
  setField morph 'scaleX' newScale
  setField morph 'scaleY' newScale

  // update costume and rotation from target
  if (isNil (costume targetM)) {updateCostume targetM true}
  setField morph 'costume' (costume targetM)
  setField morph 'costumeData' (costumeData targetM)
  setField morph 'rotation' (rotation targetM)
  if (isNil (costume morph)) {updateCostume morph}
}

method rightClicked SpriteThumbnail {
  popUpAtHand (contextMenu this) (global 'page')
  return true
}

method contextMenu SpriteThumbnail {
  menu = (menu nil this)
  targetM = (morph target)
  addItem menu 'show' 'showTarget' 'unhide and/or move instance onstage'
  addItem menu 'come to front' (action 'comeToFront' (morph target)) 'show this object on top of its siblings'
  addLine menu
  addItem menu 'explore' (action 'explore' (hand (page morph)) target) 'open an explorer window on this object''s internals'
  addLine menu
  addItem menu 'scale...' (action 'scalingHandle' target) 'scale this object'
  addItem menu 'rotate...' (action 'rotationHandle' target) 'rotate this object'
  addItem menu 'rotation point...' (action 'pinHandle' target) 'edit the point about which this object rotates'
  if (or ((pinX targetM) != 0) ((pinY targetM) != 0)) {
    addItem menu 'set rotation point to center...' (action 'setPin' targetM 0 0) 'make this object''s rotation point be its center'
  }
  if (isUserDefined (classOf (handler (owner (morph target))))) {
	if (rotateWithOwner targetM) {
	  rotationLabel = 'rotate independently'
	} else {
	  rotationLabel = 'rotate with owner'
	}
	addItem menu rotationLabel (action 'toggleRotationStyle' targetM)
	addLine menu
	addItem menu 'slide...' (action 'moveHandle' target) 'move this object within its owner'
	addItem menu 'detach' 'detachFromOwner' 'detach this object from its owner'
  }
  addLine menu
  addItem menu 'delete' (action 'destroy' (morph target) false)
  return menu
}

method showTarget SpriteThumbnail {
  container = (ownerThatIsA (morph target) 'Stage')
  if (isNil container) {container = (morph (global 'page'))}
  m = (morph target)
  keepWithin m (bounds container)
  setAlpha m 255
  show m
}

method detachFromOwner SpriteThumbnail {
  targetM = (morph target)
  container = (ownerThatIsA targetM 'Stage')
  if (isNil container) {container = (morph (global 'page'))}
  addPart container targetM
  setGrabRule targetM 'handle'
  moveBy targetM 10 10 // move a little to give feedback
}
